# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" cvo-qram """

import numpy as np
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit.library.standard_gates import UGate
from qclib.util import _compute_matrix_angles


class CVOQRAM:
    """ cvoqram """
    def __init__(self, nbits, data, with_aux=True):
        self.with_aux = with_aux
        self.initialization(nbits)
        self.circuit.x(self.aux[0])
        for k, binary_string_end_feature in enumerate(data):
            binary_string, feature = binary_string_end_feature

            control = CVOQRAM._select_controls(binary_string)
            self._flip_flop(control)
            self._load_superposition(feature, control, with_aux)
            if k < len(data) - 1:
                self._flip_flop(control)
            else:
                break

    def initialization(self, nbits):
        """ Inicialize quantum registers """
        self.aux = QuantumRegister(1, name='u')
        self.memory = QuantumRegister(nbits, name='m')

        if self.with_aux:
            self.anc = QuantumRegister(nbits-1, name='anc')
            self.circuit = QuantumCircuit(self.aux, self.anc, self.memory)
        else:
            self.circuit = QuantumCircuit(self.aux, self.memory)
        self.nbits = nbits
        self.norm = 1

    def _flip_flop(self, control):

        for k in control:
            self.circuit.cx(self.aux[0], self.memory[k])

    @staticmethod
    def _select_controls(binary_string):
        control = []
        for k, bit in enumerate(binary_string[::-1]):
            if bit == 1:
                control.append(k)
        return control



    def mcuvchain(self, alpha, beta, phi, control):
        """
         N-qubit controlled-unitary gate
        """

        lst_ctrl = control
        lst_ctrl_reversed = list(reversed(lst_ctrl))
        self.circuit.rccx(self.memory[lst_ctrl_reversed[0]],
                          self.memory[lst_ctrl_reversed[1]],
                          self.anc[self.nbits-2])

        tof = {}
        i = self.nbits-1
        for ctrl in lst_ctrl_reversed[2:]:
            self.circuit.rccx(self.anc[i-1],
                              self.memory[ctrl],
                              self.anc[i-2])
            tof[ctrl] = [i-1, i-2]
            i -= 1

        self.circuit.cu(alpha, beta, phi, 0, self.anc[i-1], self.aux[0])

        for ctrl in lst_ctrl[:-2]:
            self.circuit.rccx(self.anc[tof[ctrl][0]],
                              self.memory[ctrl],
                              self.anc[tof[ctrl][1]])

        self.circuit.rccx(self.memory[lst_ctrl[-1]],
                          self.memory[lst_ctrl[-2]],
                          self.anc[self.nbits-2])

    def _load_superposition(self, feature, control, with_aux=True):
        """
        Load pattern in superposition
        """

        theta, phi, lam = _compute_matrix_angles(feature, self.norm)

        if len(control) == 0:
            self.circuit.u(theta, phi, lam, self.aux[0])
        elif len(control) == 1:
            self.circuit.cu(theta, phi, lam, 0, self.memory[control[0]], self.aux[0])
        else:
            if with_aux:
                self.mcuvchain(theta, phi, lam, control)
            else:
                gate = UGate(theta, phi, lam).control(len(control))
                self.circuit.append(gate, self.memory[control] + [self.aux[0]])

        self.norm = self.norm - np.absolute(np.power(feature, 2))


def cvoqram_initialize(state, with_aux=True):
    """
    Creates a circuit to initialize a quantum state arXiv:

    For instance, to initialize the state a|001>+b|100>
        $ state = [('001', a), ('100', b)]
        $ circuit = sparse_initialize(state)

    Parameters
    ----------
    state: list of [(str,float)]
        A unit vector representing a quantum state.
        str: binary string
        float: amplitude

    Returns
    -------
    QuantumCircuit to initialize the state

    """
    qbit = state[0][0]
    size = len(qbit)
    n_qubits = int(size)
    memory = CVOQRAM(n_qubits, state, with_aux)
    return memory.circuit



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test majority gate """

from unittest import TestCase
from qiskit import ClassicalRegister, QuantumRegister, QuantumCircuit
from qclib.gates.majority import operate as majority
from qclib.util import get_counts

class TestMajority(TestCase):
    """ Testing qclib.gate.majority """
    @staticmethod
    def _run_majority(bin_input):
        """ Run majority gate and return counts """
        # initialize quantum registers
        controls = QuantumRegister(len(bin_input), 'controls')
        target = QuantumRegister(1, 'target')
        output = ClassicalRegister(1)
        circuit = QuantumCircuit(controls, target, output)

        # Pattern basis encoding
        for k, bit in enumerate(bin_input):
            if bit == 1:
                circuit.x(controls[k])

        majority(circuit, controls, target)

        # measure output and verify results
        # pylint: disable=maybe-no-member
        circuit.measure(target, output)
        counts = get_counts(circuit)

        return counts

    def _test_majority(self, n_qubits):
        """ Testing majority """
        for i in range(2 ** n_qubits):
            bits = [int(j) for j in '{:0{}b}'.format(i, n_qubits)]
            counts = TestMajority._run_majority(bits)

            if sum(bits) >= n_qubits/2:
                self.assertTrue(counts['1'] / 1024 == 1)
            else:
                self.assertTrue(counts['0'] / 1024 == 1)

    # Keep tests separated by number of qubits to make it easier to identify errors.
    def test_majority_2(self):
        """ Testing 2 qubits majority """
        TestMajority._test_majority(self, 2)

    def test_majority_3(self):
        """ Testing 3 qubits majority """
        TestMajority._test_majority(self, 3)

    def test_majority_4(self):
        """ Testing 4 qubits majority """
        TestMajority._test_majority(self, 4)

    def test_majority_5(self):
        """ Testing 5 qubits majority """

        TestMajority._test_majority(self, 5)

    def test_majority_6(self):
        """ Testing 6 qubits majority """
        TestMajority._test_majority(self, 6)

    def test_majority_7(self):
        """ Testing 7 qubits majority """
        TestMajority._test_majority(self, 7)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
https://arxiv.org/abs/2108.10182
"""


from graphviz    import Digraph

def is_leaf(tree):
    """
    :param tree: a tree node
    :return: True if tree is a leaf
    """
    if tree.left is None and tree.right is None:
        return True

    return False

def remove_leafs(tree):
    """ remove tree leafs """
    if tree.left:
        if is_leaf(tree.left):
            tree.left = None
        else:
            remove_leafs(tree.left)

    if tree.right:
        if is_leaf(tree.right):
            tree.right = None
        else:
            remove_leafs(tree.right)

def leftmost(tree):
    """
    :param tree: a tree node
    :return: the leftmost node relative to tree, or None if tree is leaf.
    """
    if tree.left:
        return tree.left

    return tree.right

def node_index(tree):
    """
    :param tree: a tree node
    :return: the total index of the node in the tree.
    """
    return 2**tree.level-1+tree.index

def root_node(tree, level):
    """
    :param tree: a tree node
    :param level: level of the subtree (0 for the tree root)
    :return: subtree root at level
    """
    root = tree
    while root.level > level:
        root = root.parent

    return root

def children(nodes):
    """
    Search and list all the nodes childs.
    :param nodes: a list with tree nodes
    :return: a list with nodes childs
    """
    child = []
    for node in nodes:
        if node.left:
            child.append(node.left)
        if node.right:
            child.append(node.right)

    return child

def length(tree):
    """
    Count the total number of the tree nodes.
    :param tree: a tree node
    :return: the total of nodes in the subtree
    """
    if tree:
        n_nodes = length(tree.left)
        n_nodes += length(tree.right)
        n_nodes += 1
        return n_nodes
    return 0

def level_length(tree, level):
    """
    Count the total number of the tree nodes in the level.
    :param tree: a tree node
    :param level: a tree level
    :return: the total of nodes in the subtree level
    """
    if tree:
        if tree.level < level:
            n_nodes_level = level_length(tree.left, level)
            n_nodes_level += level_length(tree.right, level)
            return n_nodes_level

        return 1

    return 0

def height(root):
    """
    Count the number of levels in the tree.
    :param root: subtree root node
    :return: the total of levels in the subtree defined by root
    """
    n_levels = 0
    left = root
    while left:
        n_levels += 1
        left = leftmost(left)

    return n_levels

def left_view(root, stop_level):
    """
    :param root: subtree root node
    :param stop_level: level below root to stop the search
    :return: list of leftmost nodes between root level and stop_level
    """
    branch = []
    left = root
    while left and left.level <= stop_level:
        branch.append(left)
        left = leftmost(left)

    return branch

def subtree_level_index(root, tree):
    """
    :param root: subtree root node
    :param tree: a tree node
    :return: the index of tree node repective to the subtree defined by root
    """
    return tree.index - root.index * 2 ** (tree.level - root.level)

def subtree_level_leftmost(root, level):
    """
    :param root: subtree root node
    :param level: level to search for the leftmost node
    :return: the leftmost tree node repective to the subtree defined by root
    """
    left = root
    while left and left.level < level:
        left = leftmost(left)
    return left

def subtree_level_nodes(tree, level, level_nodes):
    """
    Search and list all the nodes in the indicated level of the tree defined by
    the first value of tree (subtree root).
    :param tree: current tree node, starts with subtree root node
    :param level: level to search for the nodes
    :out param level_nodes: a list with the level tree nodes repective to the
                            subtree defined by root, ordered from left to right
    """
    if tree.level < level:
        if tree.left:
            subtree_level_nodes(tree.left, level, level_nodes)
        if tree.right:
            subtree_level_nodes(tree.right, level, level_nodes)
    else:
        level_nodes.append(tree)

def tree_visual_representation(tree, dot=None):
    """
    :param tree: A binary tree, with str(tree) defined
    """

    if dot is None:
        dot = Digraph()
        dot.node(str(tree))

    if tree.left:
        dot.node(str(tree.left))
        dot.edge(str(tree), str(tree.left))
        dot = tree_visual_representation(tree.left, dot=dot)

    if tree.right:
        dot.node(str(tree.right))
        dot.edge(str(tree), str(tree.right))
        dot = tree_visual_representation(tree.right, dot=dot)

    return dot



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
https://arxiv.org/abs/2108.10182
"""

from qclib.state_preparation.util.tree_utils import leftmost, children

def bottom_up(angle_tree, circuit, start_level):
    """ bottom_up state preparation """

    if angle_tree and angle_tree.level < start_level:

        if angle_tree.angle_y != 0.0:
            circuit.ry(angle_tree.angle_y, angle_tree.qubit)
            if angle_tree.angle_z != 0.0:
                circuit.rz(angle_tree.angle_z, angle_tree.qubit)

        bottom_up(angle_tree.left, circuit, start_level)
        bottom_up(angle_tree.right, circuit, start_level)

        _apply_cswaps(angle_tree, circuit)

def top_down(angle_tree, circuit, start_level, control_nodes=None, target_nodes=None):
    """ top down state preparation """
    if angle_tree:
        if angle_tree.level < start_level:
            top_down(angle_tree.left, circuit, start_level)
            top_down(angle_tree.right, circuit, start_level)
        else:
            if target_nodes is None:
                control_nodes = []                           # initialize the controls
                target_nodes = [angle_tree]                  # start by the subtree root
            else:
                target_nodes = children(target_nodes)        # all the nodes in the current level

            angles_y = [float(node.angle_y) for node in target_nodes]
            angles_z = [float(node.angle_z) for node in target_nodes]
            target_qubit = target_nodes[0].qubit
            control_qubits = [node.qubit for node in control_nodes]
            circuit.ucry(angles_y, control_qubits[::-1], target_qubit)     # qiskit reverse
            if any(angles_z) != 0.0:
                circuit.ucrz(angles_z, control_qubits[::-1], target_qubit) # qiskit reverse

            control_nodes.append(angle_tree)                 # add current node to the controls list

            # walk to the first node of the next level.
            top_down(angle_tree.left, circuit, start_level,
                     control_nodes=control_nodes, target_nodes=target_nodes)

def _apply_cswaps(angle_tree, circuit):

    if angle_tree.angle_y != 0.0:
        left = angle_tree.left
        right = angle_tree.right

        while left and right:
            circuit.cswap(angle_tree.qubit, left.qubit, right.qubit)

            left = left.left
            right = leftmost(right)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the gleining.py module.
"""

import unittest
import numpy as np
from qclib.state_preparation.gleining import initialize
from qclib.util import get_state

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

class TestGleining(unittest.TestCase):

    def test_two_states_uniform(self):
        state_vector = 1 / np.sqrt(2) * np.array([1, 0, 0, 0, 0, 1, 0, 0])
        circ = initialize(state_vector)
        state = get_state(circ)
        self.assertTrue(np.allclose(state_vector, state))

    def test_three_states_superposition(self):
        state_vector = 1 / np.sqrt(168) * np.array([0, 2, 0, 0, 8, 0, 0, 10])
        circ = initialize(state_vector)
        state = get_state(circ)
        self.assertTrue(np.allclose(state_vector, state))

    def test_three_states_uniform_superposition(self):
        state_vector = 1 / np.sqrt(3) * np.array([0, 1, 0, 0, 1, 0, 0, 1])
        circ = initialize(state_vector)
        state = get_state(circ)
        self.assertTrue(np.allclose(state_vector, state))

    def test_three_states_superposition_with_complex_features(self):
        state_vector = np.array([0, complex(np.sqrt(0.1), np.sqrt(0.1)), 0, 0,
                                 complex(np.sqrt(0.1), np.sqrt(0.2)), 0, 0, np.sqrt(0.5)])
        circ = initialize(state_vector)
        state = get_state(circ)
        self.assertTrue(np.allclose(state_vector, state))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test uniform amplitude initialize """

from unittest import TestCase
import numpy as np
from qclib.state_preparation.ventura import initialize
from qclib.util import get_state

class TestVentura(TestCase):
    """ Testing qclib.state_preparation.ventura.initialize """
    def test_binary_function(self):
        """
          Testing a function (f(z) in {0, 1, ..., N-1}) with z in {0,1}^n.
        """
        # couples of input and output.
        function_io = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 11: 0, 12: 1, 13: 2, 14: 3, 15: 4}
        nnzero = len(function_io)
        max_denominator = max(function_io.values())-1
        n_qubits = 4

        circuit = initialize(function_io, n_qubits=n_qubits, n_output_values=max_denominator)
        state = get_state(circuit)

        desired_state = []
        for k in range(2**n_qubits):
            # The amplitudes of modulus "1/sqrt(m)" will be "2 pi / N" radians
            # apart from each other on the complex plane.
            if k in function_io:
                amplitude = -1/np.sqrt(nnzero)*np.exp(function_io[k]*1j*2*np.pi/max_denominator)
                desired_state.append(amplitude)
            else:
                desired_state.append(0)

        # The data is encoded in the first 2^n amplitudes of the circuit state vector.
        self.assertTrue(np.allclose(desired_state, state[:2**n_qubits]))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
https://arxiv.org/abs/2108.10182
"""

from dataclasses import dataclass
from typing import NamedTuple
import numpy as np


class Amplitude(NamedTuple):
    """
    Named tuple for amplitudes
    """
    index: int
    amplitude: float

    def __str__(self):
        return str(self.index) + ':' + \
               '{0:.2g}'.format(self.amplitude)

@dataclass
class Node:
    """
    Binary tree node used in state_decomposition function
    """

    index: int
    level: int
    amplitude: float
    left: 'Node'
    right: 'Node'

    def __str__(self):
        return str(self.level) + '_' + \
               str(self.index) + '\n' + \
               '{0:.2g}'.format(self.amplitude)

def state_decomposition(nqubits, data):
    """
    :param nqubits: number of qubits required to generate a
                    state with the same length as the data vector (2^nqubits)
    :param data: list with exactly 2^nqubits pairs (index, amplitude)
    :return: root of the state tree
    """
    new_nodes = []

    # leafs
    for k in data:
        new_nodes.append(Node(k.index, nqubits, k.amplitude, None, None))

    # build state tree
    while nqubits > 0:
        nodes = new_nodes
        new_nodes = []
        nqubits = nqubits - 1
        k = 0
        n_nodes = len(nodes)
        while k < n_nodes:
            mag = np.sqrt(np.abs(nodes[k].amplitude) ** 2 + np.abs(nodes[k + 1].amplitude) ** 2)
            arg = (np.angle(nodes[k].amplitude) + np.angle(nodes[k + 1].amplitude)) / 2

            amp = mag * np.exp(1j*arg)

            new_nodes.append(Node(nodes[k].index // 2, nqubits,
                                  amp,
                                  nodes[k],
                                  nodes[k + 1]))
            k = k + 2

    tree_root = new_nodes[0]
    return tree_root



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the baa.py module.
"""
import sys

from test.test_baa_schmidt import TestBaaSchmidt
import datetime
import os
from multiprocessing import Pool
from unittest import TestCase

import numpy as np
import qiskit
from qiskit import QiskitError
from qiskit.circuit.random import random_circuit

from qclib.entanglement import geometric_entanglement, \
    meyer_wallach_entanglement
from qclib.state_preparation.baa_schmidt import initialize
from qclib.state_preparation.schmidt import cnot_count as schmidt_cnots
from qclib.util import get_state

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring


use_parallel = os.getenv('QLIB_TEST_PARALLEL', 'False') == 'True'


def get_vector(e_lower: float, e_upper: float, num_qubits: int,
               start_depth_multiplier=1, measure='meyer_wallach'):
    entanglement = -1.0

    if isinstance(start_depth_multiplier, int):
        multiplier = start_depth_multiplier
    elif isinstance(start_depth_multiplier, (tuple, list, np.ndarray)):
        assert len(start_depth_multiplier) > 1
        multiplier = np.random.randint(start_depth_multiplier[0], start_depth_multiplier[1])
    else:
        raise ValueError("start_depth_multiplier must be either an int or an array-like of int.")

    iteration = 0
    entanglements = []
    vector = np.ndarray(shape=(0,))
    while e_lower > entanglement or entanglement > e_upper:
        np.random.seed()
        q_circuit: qiskit.QuantumCircuit = random_circuit(num_qubits, multiplier * num_qubits)
        vector = get_state(q_circuit)
        if measure == 'geometric':
            entanglement = geometric_entanglement(vector)
        elif measure == 'meyer_wallach':
            entanglement = meyer_wallach_entanglement(vector)
        else:
            raise ValueError(f'Entanglement Measure {measure} unknown.')
        iteration += 1
        if iteration > 100:
            multiplier += 1
            iteration = 0
            if not use_parallel:
                print(
                    f'{multiplier} ({np.min(entanglements):.4f}-{np.max(entanglements):.4f})',
                    end='\n', flush=True
                )
            entanglements = []
        else:
            entanglements.append(entanglement)
            if not use_parallel:
                print('.', end='', flush=True)
    if not use_parallel:
        print(
            f'Final {multiplier} ({np.min(entanglements):.4f}-{np.max(entanglements):.4f})',
            end='\n', flush=True
        )
    return vector, entanglement, multiplier * num_qubits


def initialize_loss(fidelity_loss, state_vector=None, n_qubits=5, strategy='brute_force',
                    use_low_rank=False):

    if state_vector is None:
        state_vector = np.random.rand(2**n_qubits) + np.random.rand(2**n_qubits) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

    circuit, node = initialize(
        state_vector, max_fidelity_loss=fidelity_loss, strategy=strategy, use_low_rank=use_low_rank,
        return_node=True
    )
    state = get_state(circuit)
    fidelity = TestBaaSchmidt.fidelity(state_vector, state)

    try:
        basis_circuit = qiskit.transpile(
            circuit, basis_gates=['rx', 'ry', 'rz', 'cx'], optimization_level=0
        )
        cnots = len([d[0] for d in basis_circuit.data if d[0].name == 'cx'])
        depth = basis_circuit.depth()
    except QiskitError as ex:
        print(ex)
        return -1, -1, -1, node

    return cnots, depth, round(1 - fidelity, 4), node


def execute_experiment(exp_idx,  num_qubits, entanglement_bounds,
                       max_fidelity_losses, return_state=False,
                       start_depth_multiplier=1):
    print(f"Starting {exp_idx,  num_qubits, entanglement_bounds, max_fidelity_losses}")

    # State Generation
    state_vector, entganglement, depth = get_vector(
        *entanglement_bounds, num_qubits, start_depth_multiplier=start_depth_multiplier,
        measure='geometric'
    )
    mw_entanglement = meyer_wallach_entanglement(state_vector)
    geo_entanglement = geometric_entanglement(state_vector)
    cnots = schmidt_cnots(state_vector)
    if not use_parallel:
        print(
            f"Found state for entanglement bounds {entganglement} in {entanglement_bounds}. "
            f"State preparation needs {cnots}."
        )

    # Benchmark against real Algorithm
    real_cnots_benchmark, real_depth_benchmark, real_fidelity_loss_benchmark, _ = initialize_loss(
        state_vector=state_vector, fidelity_loss=0.0, use_low_rank=False
    )
    data_result = []
    for max_fidelity_loss in max_fidelity_losses:
        for use_low_rank in [False, True]:
            for strategy in ['brute_force', 'greedy']:
                if not use_parallel:
                    print(
                        f"[{max_fidelity_loss}] {strategy.upper()} "
                        f"{'With' if use_low_rank else 'No'} Low Rank Processing....", end=''
                    )
                start_time = datetime.datetime.now()
                real_cnots, real_depth, real_fidelity_loss, node = initialize_loss(
                    state_vector=state_vector, fidelity_loss=max_fidelity_loss,
                    use_low_rank=use_low_rank, strategy=strategy
                )
                end_time = datetime.datetime.now()
                duration = (end_time - start_time).total_seconds()
                vector_data = list(
                    zip(node.ranks, [list(v.shape) for v in node.vectors])
                )
                data = [
                    exp_idx, use_low_rank, strategy, num_qubits, depth, cnots, geo_entanglement,
                    mw_entanglement, max_fidelity_loss, node.total_saved_cnots,
                    node.total_fidelity_loss, vector_data, real_cnots, real_cnots_benchmark,
                    real_depth, real_depth_benchmark, real_fidelity_loss,
                    real_fidelity_loss_benchmark, duration
                ]
                if not use_parallel:
                    print(f"in {duration} secs")
                data_result.append(data)

    # Experiment transcription
    data_result = np.asarray(data_result, dtype=object)
    print(f"Done {exp_idx,  num_qubits, entanglement_bounds, max_fidelity_losses}")
    if return_state:
        return data_result, state_vector
    return data_result


class TestBaa(TestCase):
    columns = [
        'id', 'with_low_rank', 'strategy', 'num_qubits', 'depth', 'cnots', 'entganglement',
        'entganglement (MW)', 'max_fidelity_loss', 'total_saved_cnots', 'total_fidelity_loss',
        'data', 'real_cnots', 'real_cnots_benchmark', 'real_depth', 'real_depth_no_approx',
        'real_fidelity_loss', 'real_fidelity_loss_benchmark', 'duration',
        'predicted_cnots', 'benchmark_fidelity_loss_pass', 'approximation_calculation_pass',
        'real_approximation_calculation_pass'
    ]

    def test(self):
        num_qubits = 7
        entanglement_bounds = (0.4, 1.0)
        max_fidelity_loss = np.linspace(0.1, 1.0, 4)
        number_of_experiments = 5

        data = [
            (i, num_qubits, entanglement_bounds, max_fidelity_loss)
            for i in range(number_of_experiments)
        ]
        if use_parallel:
            with Pool() as pool:
                result = pool.starmap(execute_experiment, data)
        else:
            result = [execute_experiment(*d) for d in data]

        test_data = np.concatenate(result)
        # Predicted CNOTs
        test_data = np.hstack([
            test_data,
            (test_data[:, self.columns.index('cnots')]
             - test_data[:, self.columns.index('total_saved_cnots')]).reshape(-1, 1)
        ])

        # Tests:
        # Calculation Tests
        # The benchmark will not change the state at all, so it must be
        # essentially zero
        test_data = np.hstack([
            test_data,
            (np.abs(test_data[:, self.columns.index('real_fidelity_loss_benchmark')]) < 1e-6
             ).reshape(-1, 1)
        ])
        # The expected / predicted fidelity loss must be less or equal to the
        # max fidelity loss
        test_data = np.hstack([
            test_data,
            (test_data[:, self.columns.index('max_fidelity_loss')]
             >= test_data[:, self.columns.index('total_fidelity_loss')]
             ).reshape(-1, 1)
        ])

        # The real Tests
        # The real measured fidelity measure must be less or equal to the
        # configured mx fidelity loss
        test_data = np.hstack([
            test_data,
            (test_data[:, self.columns.index('real_fidelity_loss')]
             - test_data[:, self.columns.index('max_fidelity_loss')] < 0.1
             ).reshape(-1, 1)
        ])

        # Attention: this is a set of tests that tests how well the CNOT-estimation works
        # However, we do not test this here. As a result, also the prediciton of CNOTs cannot
        # reliably be tested.
        # START: COMMENTED OUT AND LEFT FOR DOCUMENTATION REASONS
        # # The predicted maximum CNOT gates and the no-approximation count
        # # should be within 10%
        # test_data['cnot_prediction_calculation_pass'] = (
        #         np.abs(test_data['real_cnots_benchmark']
        #         - test_data['cnots']) < 0.2 * test_data['cnots']
        # )
        # # The predicted CNOT gates should be within an error margin of 10%
        # test_data['saved_cnots_calculation_pass'] = (
        #         np.abs(test_data['predicted_cnots']
        #         - test_data['real_cnots']) <= 0.2 * (test_data['predicted_cnots'])
        # )
        # END: COMMENTED OUT AND LEFT FOR DOCUMENTATION REASONS

        test_passed = True
        total_experiments = number_of_experiments * max_fidelity_loss.shape[0] * 4
        fails_to_still_pass = int(np.ceil(total_experiments * 0.01))
        benchmark_fidelity_loss_fail_count = (
                test_data.shape[0]
                - test_data[
                    test_data[:, self.columns.index('benchmark_fidelity_loss_pass')].astype(bool)
                ].shape[0]
        )
        approximation_calculation_fail_count = (
                test_data.shape[0]
                - test_data[
                    test_data[:, self.columns.index('approximation_calculation_pass')].astype(bool)
                ].shape[0]
        )
        real_approximation_calculation_fail_count = (
                test_data.shape[0] - test_data[
                test_data[:, self.columns.index('real_approximation_calculation_pass')].astype(bool)
            ].shape[0]
        )
        if benchmark_fidelity_loss_fail_count > 0:
            print(f"[WARNING] NOT ALL benchmark_fidelity_loss_pass are true ({benchmark_fidelity_loss_fail_count})",
                  file=sys.stderr)
        if benchmark_fidelity_loss_fail_count > fails_to_still_pass:
            print("[FAIL] benchmark_fidelity_loss_pass must be true", file=sys.stderr)
            test_passed = False
        if approximation_calculation_fail_count > 0:
            print(f"[WARNING] NOT ALL approximation_calculation_pass are true ({approximation_calculation_fail_count})",
                  file=sys.stderr)
        if approximation_calculation_fail_count > fails_to_still_pass:
            print("[FAIL] approximation_calculation_pass must be true", file=sys.stderr)
            test_passed = False
        if real_approximation_calculation_fail_count > 0:
            print(f"[WARNING] NOT ALL real_approximation_calculation_pass are true "
                  f"({real_approximation_calculation_fail_count})", file=sys.stderr)
        if real_approximation_calculation_fail_count > fails_to_still_pass:
            print("[FAIL] real_approximation_calculation_pass must be true", file=sys.stderr)
            test_passed = False

        import importlib.util
        pandas_loader = importlib.util.find_spec('pandas')
        if pandas_loader is None:
            with np.printoptions(precision=2, linewidth=1000, suppress=True, threshold=sys.maxsize, floatmode='fixed'):
                print(test_data)
        else:
            import pandas as pd
            print(pd.DataFrame(test_data, columns=self.columns).to_string())

        self.assertTrue(test_passed, 'The tests should all pass.')

    def test_no_ops(self):
        # The Test is based on randomly generated states. They all should
        # create the correct fidelity (1.0). After 10 attempts we may well
        # find one that fails. If it doesn't it is probably okay.
        for lower_bound_int in range(10):
            state_vector, _, depth = get_vector(
                lower_bound_int/10, 1.0, 8, 1, measure='geometric'
            )
            cnots, depth, fidelity_loss, _ = initialize_loss(0.0, state_vector)
            if cnots == depth == fidelity_loss == -1:
                continue
            self.assertAlmostEqual(0.0, fidelity_loss, 4)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions to compute entanglement measures.
"""
from typing import Union, Tuple, List

import numpy as np
import tensorly as tl
from tensorly.tucker_tensor import TuckerTensor
from tensorly.decomposition import tucker


def _get_iota(qubit_idx: int, qubits: int, selector_bit: int, basis_state: int):
    assert selector_bit in [0, 1]
    full_mask = 2**qubits - 1

    mask_j = 1 << qubit_idx
    value = (mask_j & basis_state) >> qubit_idx

    low_mask = full_mask >> (qubits - qubit_idx)
    high_mask = full_mask & (full_mask << (qubit_idx + 1))
    new_basis_state = ((basis_state & high_mask) >> 1) + (basis_state & low_mask)

    return value == selector_bit, new_basis_state


def generalized_cross_product(vector_u: np.ndarray, vector_v: np.ndarray) -> np.ndarray:
    """
    Calculates the generalized cross product (see Eqn. (3) in quant-ph/0305094)

    Args:
        vector_u (array-like):
            The first vector (called u)
        vector_v (array-like):
            The first vector (called u)

    Returns:
        array-like: the resulting vector

    """
    entries = []
    for j in range(vector_u.shape[0]):
        for i in range(j):
            entry = np.abs(vector_u[i] * vector_v[j] - vector_u[j] * vector_v[i])**2
            entries.append(entry)
    return np.sum(entries)


def meyer_wallach_entanglement(vector: np.ndarray) -> float:
    """

    Computes the Meyer-Wallach entanglement (1,2) of a quantum state.

    [1] Meyer, D. A. & Wallach, N. R. Global entanglement in multiparticle systems. J Math Phys 43,
        4273–4278 (2002).
    [2] Brennen, G. An observable measure of entanglement for pure states of multi-qubit systems.
        P Soc Photo-opt Ins 3, 619–626 (2003).

    Args:
        vector (array-like):
            The vector of the quantum state (in computational basis)
    Returns:
        float: the entanglement which is between 0 and 1 (highest is 1)

    """
    num_qb = _to_qubits(vector.shape[0])
    meyer_wallach_entry = np.zeros(shape=(num_qb, 1))
    for j in range(num_qb):
        psi_0 = np.zeros(shape=(vector.shape[0]//2, 1), dtype=complex)
        psi_1 = np.zeros(shape=(vector.shape[0]//2, 1), dtype=complex)
        for basis_state, entry in enumerate(vector):
            delta_0, new_basis_state_0 = _get_iota(j, num_qb, 0, basis_state)
            delta_1, new_basis_state_1 = _get_iota(j, num_qb, 1, basis_state)

            if delta_0:
                psi_0[new_basis_state_0] = entry
            if delta_1:
                psi_1[new_basis_state_1] = entry

        entry = generalized_cross_product(psi_0, psi_1)
        meyer_wallach_entry[j] = entry

    return np.sum(meyer_wallach_entry) * (4/num_qb)


def geometric_entanglement(state_vector: List[complex], return_product_state=False
                           ) -> Union[float, Tuple[float, List[np.ndarray]]]:
    """

    Computes the geometric entanglement (1,2) of a quantum state.

    [1] SHIMONY, A. Degree of Entanglementa. Ann Ny Acad Sci 755, 675–679 (1995).

    [2] Barnum, H. & Linden, N. Monotones and invariants for multi-particle quantum states.
        J Phys Math Gen 34, 6787 (2001).

    Args:
        state_vector (array-like):
            The vector of the quantum state (in computational basis)

        return_product_state (bool):
            If True, return the list of product states too.

    Returns:
        float or Tuple[float, List[array-like]]: #
            the entanglement which is between 0 and 1 (highest is 1).
            If return_product_state == True, returns a tuple with a list of product state vectors.

    """
    shape = tuple([2] * _to_qubits(len(state_vector)))
    rank = [1] * _to_qubits(len(state_vector))
    tensor = tl.tensor(state_vector).reshape(shape)
    results = {}
    # The Tucker decomposition is actually a randomized algorithm.
    # We take four shots and take the min of it.
    for _ in range(4):
        decomp_tensor: TuckerTensor = tucker(
            tensor, rank=rank, verbose=False, svd='numpy_svd', init='random'
        )
        fidelity_loss = 1 - np.linalg.norm(decomp_tensor.core) ** 2
        results[fidelity_loss] = decomp_tensor

    min_fidelity_loss = min(results)

    if return_product_state:
        return min_fidelity_loss, [f.flatten() for f in results[min_fidelity_loss].factors]

    return min_fidelity_loss


def _to_qubits(n_state_vector):
    return int(np.ceil(np.log2(n_state_vector))) if n_state_vector > 0 else 0



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test creation of quantum circuits from matrices """

from unittest import TestCase
from scipy.stats import unitary_group
import numpy as np
import qiskit
from qclib.unitary import unitary, _compute_gates
from qclib.util import get_state

class TestUnitary(TestCase):
    """ Testing qclib.unitary """
    def test_unitary_csd_2qubits(self):
        """ Testing qclib.unitary with 2 qubits gate"""
        unitary_matrix = unitary_group.rvs(4)
        gate = unitary(unitary_matrix)
        state = get_state(gate)

        self.assertTrue(np.allclose(unitary_matrix[:, 0], state))

    def test_unitary_csd_5qubits(self):
        """ Testing qclib.unitary 5 qubits gate"""
        unitary_matrix = unitary_group.rvs(32)
        gate = unitary(unitary_matrix)
        state = get_state(gate)
        self.assertTrue(np.allclose(unitary_matrix[:, 0], state))

    def test_unitary_qsd_4qubits(self):
        """ Testing qclib.unitary 4 qubits gate qsd"""
        unitary_matrix = unitary_group.rvs(16)
        gate = unitary(unitary_matrix, 'qsd')
        state = get_state(gate)
        self.assertTrue(np.allclose(unitary_matrix[:, 0], state))

        circuit = qiskit.QuantumCircuit(4)
        circuit.x(0)
        circuit.append(gate, circuit.qubits)
        state = get_state(circuit)
        self.assertTrue(np.allclose(unitary_matrix[:, 1], state))

        circuit = qiskit.QuantumCircuit(4)
        circuit.x(1)
        circuit.append(gate, circuit.qubits)
        state = get_state(circuit)
        self.assertTrue(np.allclose(unitary_matrix[:, 2], state))

        circuit = qiskit.QuantumCircuit(4)

        circuit.x([0, 1, 2, 3])
        circuit.append(gate, circuit.qubits)
        state = get_state(circuit)
        self.assertTrue(np.allclose(unitary_matrix[:, 15], state))

    def test_unitary_qsd_5qubits(self):
        """ Testing qclib.unitary 5 qubits gate qsd"""
        unitary_matrix = unitary_group.rvs(16)
        gate = unitary(unitary_matrix, 'qsd')
        state = get_state(gate)
        transpiled_circuit = qiskit.transpile(gate, basis_gates=['u', 'cx'])
        n_cx = transpiled_circuit.count_ops()['cx']
        self.assertTrue(n_cx <= 120)
        self.assertTrue(np.allclose(unitary_matrix[:, 0], state))

    def test_compute_gates(self):
        """ test auxiliar funciont compute gates"""
        gate1 = unitary_group.rvs(8)
        gate2 = unitary_group.rvs(8)

        diag, v_gate, w_gate = _compute_gates(gate1, gate2)
        calc1 = v_gate @ np.diag(diag) @ w_gate
        calc2 = v_gate @ np.diag(diag).conj().T @ w_gate
        self.assertTrue(np.allclose(calc1, gate1))
        self.assertTrue(np.allclose(calc2, gate2))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test divide-and-conquer state preparation """

from unittest import TestCase
import numpy as np
from qiskit import ClassicalRegister
from qiskit.providers.aer.backends import AerSimulator
from qclib.state_preparation.dcsp import initialize
from .util import measurement

backend = AerSimulator()
SHOTS = 8192

class TestInitialize(TestCase):
    """ Testing divide-and-conquer state preparation """

    @staticmethod
    def dcsp_experiment(state):
        """ Run divide-and-conquer state preparation """
        circuit = initialize(state)

        n_qubits = int(np.log2(len(state)))
        classical_register = ClassicalRegister(n_qubits)
        circuit.add_register(classical_register)

        return measurement(circuit, n_qubits, classical_register, backend, SHOTS)

    def test_dcsp(self):
        """ Testing divide-and-conquer state preparation """
        vector = np.random.rand(16) + np.random.rand(16) * 1j
        vector = vector / np.linalg.norm(vector)

        state = TestInitialize.dcsp_experiment(vector)

        self.assertTrue(np.allclose( np.power(np.abs(vector),2), state, rtol=1e-01, atol=0.005))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test qclib.gate.mc_gate """

from unittest import TestCase

import numpy as np
from scipy.stats import unitary_group
import qiskit
import qclib.util
from qclib.gates.mc_gate import mc_gate


class TestLinearToffoli(TestCase):
    """ Testing qclib.gate.toffoli """

    def test_controlled_gate(self):
        """ Testing multi controlled gate """

        gate_u = unitary_group.rvs(2)

        circuit = qiskit.QuantumCircuit(5)
        circuit.x(1)
        circuit.x(2)
        circuit.x(3)
        circuit.x(4)

        mc_gate(gate_u, circuit, [4, 3, 2, 1], 0)

        state = qclib.util.get_state(circuit)
        self.assertTrue(np.isclose(state[30], gate_u[0,0]))
        self.assertTrue(np.isclose(state[31], gate_u[1, 0]))

        circuit2 = qiskit.QuantumCircuit(5)
        circuit2.x(0)
        circuit2.x(1)
        circuit2.x(2)
        circuit2.x(3)
        circuit2.x(4)

        mc_gate(gate_u, circuit2, [4, 3, 2, 1], 0)

        state = qclib.util.get_state(circuit2)
        self.assertTrue(np.isclose(state[30], gate_u[0, 1]))
        self.assertTrue(np.isclose(state[31], gate_u[1, 1]))

    def test_linear_toffoli3(self):
        """ Testing Toffoli control 111"""
        gate_x = np.array([[0, 1], [1, 0]])
        circuit = qiskit.QuantumCircuit(4)

        circuit.x(1)
        circuit.x(2)
        circuit.x(3)
        circuit.x(0)

        mc_gate(gate_x, circuit, [0, 1, 2], 3)

        state = qclib.util.get_state(circuit)
        exp_state = np.zeros(16, dtype=complex)
        exp_state[7] = 1

        self.assertTrue(np.allclose(state, exp_state))

    def test_linear_toffoli2(self):
        """ Testing Toffoli control 110"""
        gate_x = np.array([[0, 1], [1, 0]])

        circuit2 = qiskit.QuantumCircuit(4)
        circuit2.x(2)
        circuit2.x(3)
        circuit2.x(0)
        state1 = qclib.util.get_state(circuit2)

        circuit = qiskit.QuantumCircuit(4)
        mc_gate(gate_x, circuit, [3, 2, 1], 0)

        circuit2.compose(circuit, circuit2.qubits)
        state2 = qclib.util.get_state(circuit2)

        self.assertTrue(np.allclose(state1, state2))

    def test_linear_toffoli1(self):
        """ Testing Toffoli control 100"""
        gate_x = np.array([[0, 1], [1, 0]])

        circuit2 = qiskit.QuantumCircuit(4)
        circuit2.x(2)

        state1 = qclib.util.get_state(circuit2)

        circuit = qiskit.QuantumCircuit(4)

        mc_gate(gate_x, circuit, [0, 1, 2], 3)
        circuit2.compose(circuit, circuit2.qubits)

        state2 = qclib.util.get_state(circuit2)

        self.assertTrue(np.allclose(state1, state2))

    def test_linear_toffoli0(self):
        """ Testing Toffoli control 000"""
        gate_x = np.array([[0, 1], [1, 0]])

        circuit2 = qiskit.QuantumCircuit(4)

        state1 = qclib.util.get_state(circuit2)

        circuit = qiskit.QuantumCircuit(4)

        mc_gate(gate_x, circuit, [1, 2, 3], 0)
        circuit.compose(circuit2, circuit.qubits)

        state2 = qclib.util.get_state(circuit)

        self.assertTrue(np.allclose(state1, state2))

    def test_mct_toffoli(self):
        """ compare qiskit.mct and toffoli depth with 7 qubits """
        gate_x = np.array([[0, 1], [1, 0]])
        qcirc1 = qiskit.QuantumCircuit(6)
        qcirc1.mct([0, 1, 2, 3, 4], 5)
        t_qcirc1 = qiskit.transpile(qcirc1, basis_gates=['u', 'cx'])

        qcirc2 = qiskit.QuantumCircuit(6)
        mc_gate(gate_x, qcirc2, [0, 1, 2, 3, 4], 5)
        t_qcirc2 = qiskit.transpile(qcirc2, basis_gates=['u', 'cx'])

        self.assertTrue(t_qcirc2.depth() < t_qcirc1.depth())



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implements the decomposition of isometries using the methods
defined at https://arxiv.org/abs/1501.06911.
"""

from math import log2
import numpy as np
import scipy
import qiskit.quantum_info as qi
from qiskit import QuantumCircuit, QuantumRegister, transpile
from qiskit.extensions.quantum_initializer.uc import UCGate

# pylint: disable=maybe-no-member

def decompose(isometry, scheme='ccd'):
    """
    Decompose an isometry from m to n qubits.
    In particular, it decomposes unitaries on n qubits (m=n) or prepare a
    quantum state on n qubits (m=0).
    `https://arxiv.org/abs/1501.06911`_.
    Args:
        isometry (list): an isometry from m to n qubits (n>=2 and m<=n), i.e., a
            complex 2^n×2^m array with orthonormal columns.
        scheme (str): method to decompose the isometry ('knill', 'ccd', 'csd').
            Default is scheme='ccd'.
    Returns:
        QuantumCircuit: a quantum circuit with the isometry attached.
    Raises:
        ValueError: if isometry is not valid.
    """

    iso = isometry.astype(complex)
    if len(iso.shape) == 1:
        iso = iso.reshape(iso.shape[0], 1) # change a row vector to a column
                                           # vector (in the case of state
                                           # preparation).
    lines = iso.shape[0]
    cols = iso.shape[1]
    log_lines = log2(lines)
    log_cols = log2(cols)

    _check_isometry(iso, log_lines, log_cols)

    log_lines = int(log_lines)
    log_cols = int(log_cols)

    if scheme == 'csd':
        pass
    if scheme == 'ccd':
        return _ccd(iso, log_lines, log_cols)

    return _knill(iso, log_lines, log_cols)



#   General



def _check_isometry(iso, log_lines, log_cols):
    if not log_lines.is_integer() or log_lines < 0:
        raise ValueError(
            "The number of rows of the isometry is not a non negative power of 2."
        )
    if not log_cols.is_integer() or log_cols < 0:
        raise ValueError(
            "The number of columns of the isometry is not a non negative power of 2."
        )
    if log_cols > log_lines:
        raise ValueError(
            "The input matrix has more columns than rows."
        )
    if not _is_isometry(iso, log_cols):
        raise ValueError(
            "The input matrix has non orthonormal columns."
        )

def _is_isometry(iso, log_cols):
    identity = np.conj(iso.T).dot(iso)
    return np.allclose(identity, np.eye(int(2**log_cols)))



#   Knill



def _knill(iso, log_lines, log_cols):
    if log_lines < 2:
        raise ValueError(
            "Knill decomposition does not work on a 1 qubit isometry (N=2)."
        )

    unitary = _extend_to_unitary(iso, log_lines, log_cols)

    eigval, eigvec = np.linalg.eig(unitary)
    arg = np.angle(eigval)

    reg = QuantumRegister(log_lines)
    circuit = QuantumCircuit(reg)

    from qclib.state_preparation.schmidt import initialize as schmidt # pylint: disable=import-outside-toplevel

    for i in range(2**log_lines):                            # The eigenvalues are not necessarily
                                                             # ordered.
        if np.abs(arg[i]) > 10**-15:
            state = eigvec[:,i]

            circuit.compose( schmidt(state).inverse(), reg, inplace=True )

            circuit.x(list(range(log_lines)))
            circuit.mcp(arg[i], list(range(log_lines-1)), log_lines-1)
            circuit.x(list(range(log_lines)))

            circuit.compose( schmidt(state), reg, inplace=True )

    return circuit

def _extend_to_unitary(iso, log_lines, log_cols):
    if log_lines == log_cols:      # The isometry v is already unitary.
        unitary = iso
    else:                          # The isometry v is extended to a unitary maximizing
                                   # the numbers of eigenvalues with complex argument equal
                                   # to zero.
        null_space = np.conj(scipy.linalg.null_space(iso.T)) # The complex conjugate of the null
                                                             # space is the transformation that
                                                             # generates the state |v> from |0>.
                                                             # V=UI_{2^n,2^m} => U^-1 V=I_{2^n,2^m}
                                                             # The transposition was removed
                                                             # because it would be nullified in
                                                             # U[:, M:] = W.T .
        unitary = np.zeros((2**log_lines, 2**log_lines), dtype=complex)
        unitary[:, :2**log_cols] = iso
        unitary[:, 2**log_cols:] = null_space                # The transposition was removed
                                                             # because it nullified the
                                                             # transposition of the conjugate
                                                             # complex above (which was removed).
    return unitary



#   Column-by-column



def _ccd(iso, log_lines, log_cols):
    reg = QuantumRegister(log_lines)
    circuit = QuantumCircuit(reg)

    for k in range(2**log_cols): # iteration through columns (starting with the second column k=1).
        g_k = _g_k(iso, log_lines, k)
        g_k.name = 'G'+str(k)

        circuit.append( g_k, reg )

    if log_cols > 0:             # It clears the phases, as explained in the last sentence of the
                                 # first column on page 17.
        phases = np.angle( np.diagonal( iso[:2**log_cols,:2**log_cols] ) )
        diag = np.exp( -1j*phases )
        circuit.diagonal( diag.tolist(), list(range(log_cols)) )

    return circuit.inverse()

def _g_k(iso, log_lines, col_index):
    g_k = QuantumCircuit(log_lines)                   # Gate G columns index k, to be created.
    k_bin = '{:0{}b}'.format(col_index, log_lines)    # Binary representation of column index k.
    for i in range(log_lines):                        # G_k's subgate bit index i (s in the paper).
        target = log_lines - i - 1
        control = list(range(target))
        ancilla = list(range(target+1, log_lines))

        if _k_s(col_index, i) == 0 and _b(col_index, i+1) != 0:      # Condition defined in the
                                                                     # first paragraph of the
                                                                     # second column on page 16.
            unitary = _mc_unitary(iso, col_index, i)                 # Generates single-qubit gate
                                                                     # matrix for the
                                                                     # multicontrolled operation.
            mcg = _mc_gate(unitary, log_lines, control+ancilla, target, k_bin)
            mcg = mcg.reverse_bits()                                 # Qiskit little-endian.
            _update_isometry(iso, mcg)                               # Updates the isometry matrix.
            g_k.compose( mcg, list(range(log_lines)), inplace=True ) # Append "mcg" to the
                                                                     # operator "G_k".

        unitaries = _uc_unitaries(iso, log_lines, col_index, i)      # Generates single-qubit gates
                                                                     # matrices for the uniformly
                                                                     # controlled operation.
        ucg = _uc_gate(unitaries, log_lines, control, target)
        ucg = ucg.reverse_bits()                                     # Qiskit little-endian.
        _update_isometry(iso, ucg)                                   # Updates the isometry matrix.
        g_k.compose( ucg, list(range(log_lines)), inplace=True )     # Append "ucg" to the
                                                                     # operator "G_k".
    return g_k

def _mc_gate(unitary, n_qubits, control, target, k_bin):
    gate = QuantumCircuit(n_qubits)

    controls = []
    for i in control:
        if k_bin[i] == '1':
            controls.append(i)

    unitaries = [np.identity(2) for i in range(2**len(controls))]
    unitaries[-1] = unitary
    ucg = UCGate(unitaries, up_to_diagonal=True)
    gate.append( ucg, [target] + controls[::-1])

    return gate

def _uc_gate(unitaries, n_qubits, control, target):
    gate = QuantumCircuit(n_qubits)

    if len(control) > 0 :
        gate.uc(unitaries, control[::-1], target, up_to_diagonal=True) # "control" is reversed due
                                                                       # to UCGate implementation.
    else:
        gate.unitary(unitaries[0], target)                             # UCGate does not work with
                                                                       # target only.
    return gate

def _update_isometry(iso, gate):
    gate_matrix = qi.Operator(gate).data # "gate" matrix representation.
    iso[:,:] = gate_matrix @ iso         # Updates isometry.

def _mc_unitary(iso, col_index, bit_index):
    col = iso[:,col_index]
    idx1 =  2*_a(col_index, bit_index+1)    * 2**bit_index + _b(col_index, bit_index+1)
    idx2 = (2*_a(col_index, bit_index+1)+1) * 2**bit_index + _b(col_index, bit_index+1)

    return _unitary([ [ col[idx1] ], [ col[idx2] ] ], basis=0)

def _uc_unitaries(iso, n_qubits, col_index, bit_index):
    start = _a(col_index, bit_index+1) + 1
    if _b(col_index, bit_index+1) == 0:
        start = _a(col_index, bit_index+1)

    gates = []
    for i in range(start):
        gates.append( np.identity( 2 ) )

    col = iso[:, col_index]
    for i in range(start, 2**(n_qubits-bit_index-1)):
        idx1 =  2*i    * 2**bit_index + _b(col_index, bit_index)
        idx2 = (2*i+1) * 2**bit_index + _b(col_index, bit_index)

        gates.append( _unitary([ [ col[idx1] ], [ col[idx2] ] ], basis=_k_s(col_index, bit_index)) )

    return gates

def _unitary(iso, basis=0): #  Lemma2 of https://arxiv.org/abs/1501.06911
    iden = np.identity(2)
    iso_norm = np.linalg.norm(iso, axis=0)[0]

    if iso_norm != 0.0:
        psi = iso / iso_norm

        psi_dagger = np.conj(psi.T)

        phi = -psi_dagger.dot( iden[:,[1]] ) * \
                    iden[:,[0]] + psi_dagger.dot( iden[:,[0]] ) * iden[:,[1]]
        phi_dagger = np.conj(phi.T)

        unitary = np.kron( iden[:,[basis]], psi_dagger ) + \
                    np.kron( iden[:,[-(basis+1)]], phi_dagger)

        return unitary

    return iden

def _a(col_index, bit_index):              # col_index >> bit_index.
    return col_index // 2**bit_index       # Returns int representing n-bit_index most
                                           # significant bits.
def _b(col_index, bit_index):              # col_index ^ ((col_index >> bit_index) << bit_index)
    return col_index - (_a(col_index, bit_index) * 2**bit_index) # Returns int representing
                                                                 # bit_index less significant bits.

def _k_s(col_index, bit_index):
    return int((col_index & 2**bit_index) / 2**bit_index) # Returns the bit value at bit_index
                                                          # of col_index (k in the paper).



# CNOT count



def cnot_count(isometry, scheme='ccd', method='estimate'):
    """
    Count the number of CNOTs to decompose the isometry.
    """
    if method == 'estimate':
        return _cnot_count_estimate(isometry, scheme)

    # Exact count
    circuit = decompose(isometry, scheme)
    transpiled_circuit = transpile(circuit, basis_gates=['u1','u2','u3','cx'],
                                                            optimization_level=0)
    count_ops = transpiled_circuit.count_ops()
    if 'cx' in count_ops:
        return count_ops['cx']

    return 0

def _cnot_count_estimate(isometry, scheme='ccd'):
    """
    Estimate the number of CNOTs to decompose the isometry.
    """
    iso = isometry.astype(complex)
    if len(iso.shape) == 1:
        iso = iso.reshape(iso.shape[0], 1)

    log_lines = int(log2(iso.shape[0]))
    log_cols = int(log2(iso.shape[1]))

    if scheme == 'knill':
        return _cnot_count_estimate_knill(isometry, log_lines, log_cols)

    # CCD
    return _cnot_count_estimate_ccd(log_lines, log_cols)

def _cnot_count_estimate_knill(iso, log_lines, log_cols):
    """
    Estimate the number of CNOTs to decompose the isometry using Knill.
    """
    unitary = _extend_to_unitary(iso, log_lines, log_cols)

    eigval, eigvec = np.linalg.eig(unitary)
    arg = np.angle(eigval)

    from qclib.state_preparation.schmidt import cnot_count as schmidt_cnot_count # pylint: disable=import-outside-toplevel

    cnots = 0
    for i in range(2**log_lines):
        if np.abs(arg[i]) > 10**-15:
            state = eigvec[:,i]

            # Two times Schmidt state preparation
            cnots += 2 * schmidt_cnot_count(state)

            # MCP
            if log_lines == 2:
                cnots += 1
            elif log_lines > 2:
                cnots += 16*log_lines**2 - 60*log_lines + 42

    return cnots

def _cnot_count_estimate_ccd(log_lines, log_cols):
    """
    Estimate the number of CNOTs to decompose the isometry using CCD.
    """
    cnots = 0
    for k in range(2**log_cols):
        k_bin = '{:0{}b}'.format(k, log_lines)
        # G_K
        for i in range(log_lines):
            target = log_lines - i - 1
            control = list(range(target))
            ancilla = list(range(target+1, log_lines))

            if _k_s(k, i) == 0 and _b(k, i+1) != 0:
                # MCG implemented as a UCG up to a diagonal
                n_qubits = sum([k_bin[q] == '1' for q in control+ancilla]) + 1
                cnots += 2**(n_qubits - 1) - 1

            # UCG up to a diagonal
            n_qubits = len(control) + 1
            cnots += 2**(n_qubits - 1) - 1

    # Diagonal
    if log_cols > 0:
        cnots += 2**log_cols - 2

    return cnots



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" https://arxiv.org/abs/2011.07977 """

import numpy as np
from qiskit import QuantumCircuit, QuantumRegister
from qclib.util import _compute_matrix_angles

class CVQRAM:
    """ https://arxiv.org/abs/2011.07977 """
    def __init__(self, nbits, data, mode='v-chain'):

        self.initialization(nbits, mode)
        norm = 1
        if mode == 'v-chain':
            self.circuit.x(self.qr_u2[0])
        elif mode == 'mct':
            self.circuit.x(self.qr_u0[1])

        control = range(self.memory.size)
        for binary_string, amplitude in data:

            self._load_binary(binary_string, mode)
            self.load_superposition(amplitude, mode, norm, control)
            self._load_binary(binary_string, mode)

    def initialization(self, nbits, mode):
        """ Initialize quantum registers"""

        self.nbits = nbits

        self.memory = QuantumRegister(self.nbits, name='m')


        if mode=='mct':
            self.qr_u0 = QuantumRegister(2, name='u0')
            self.circuit = QuantumCircuit(self.qr_u0, self.memory)
        elif mode=='v-chain':
            self.aux = QuantumRegister(nbits-1, name='anc')
            self.qr_u1 = QuantumRegister(1, name='u1')
            self.qr_u2 = QuantumRegister(1, name='u2')
            self.circuit = QuantumCircuit(self.qr_u1, self.qr_u2, self.memory, self.aux, )


    def _load_binary(self, binary_string, mode):

        for bit_index, bit in enumerate(binary_string):

            if bit == '1':
                if mode=='v-chain':
                    self.circuit.cx(self.qr_u1[0], self.memory[bit_index])
                elif mode=='mct':
                    self.circuit.cx(self.qr_u1[1], self.memory[bit_index])
            elif bit == '0':
                self.circuit.x(self.memory[bit_index])


    def load_superposition(self, feature, mode, norm, control):
        """
        Load pattern in superposition
        """

        alpha, beta, phi = _compute_matrix_angles(feature, norm)

        if mode =='v-chain':

            self.circuit.rccx(self.memory[control[0]],
                              self.memory[control[1]], self.aux[0])

            for j in range(2, len(control)):
                self.circuit.rccx(self.memory[control[j]], self.aux[j - 2], self.aux[j - 1])

            self.circuit.cx(self.aux[len(control) - 2], self.qr_u1[0])

            self.circuit.cu3(alpha, beta, phi, self.qr_u1[0], self.qr_u2[0])

            self.circuit.cx(self.aux[len(control) - 2], self.qr_u1[0])

            for j in reversed(range(2, len(control))):
                self.circuit.rccx(self.memory[control[j]], self.aux[j - 2], self.aux[j - 1])

            self.circuit.rccx(self.memory[control[0]],
                              self.memory[control[1]], self.aux[0])

        if mode =='mct':
            self.circuit.mct(self.memory, self.qr_u0[0])
            self.circuit.cu3(alpha, beta, phi, self.qr_u0[0], self.qr_u0[1])
            self.circuit.mct(self.memory, self.qr_u0[0])
        norm = norm - np.absolute(np.power(feature, 2))

def cvqram_initialize(state):
    """
    Creates a circuit to initialize a quantum state arXiv:2011.07977
    """
    qbit = state[0][0]
    size = len(qbit)
    n_qubits = int(size)
    memory = CVQRAM(n_qubits, state)
    return memory.circuit



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Bounded Approximation Algorithm.
https://arxiv.org/abs/2111.03132
"""

from dataclasses import dataclass
from itertools import combinations, chain
from typing import List, Optional, Tuple
from math import log2, sqrt
import numpy as np
import tensorly as ty

from qclib.entanglement import geometric_entanglement
from qclib.state_preparation.schmidt import cnot_count as schmidt_cnots, \
                                            schmidt_decomposition, \
                                            schmidt_composition, \
                                            low_rank_approximation

# pylint: disable=missing-class-docstring

def adaptive_approximation(state_vector, max_fidelity_loss, strategy='greedy',
                                        max_combination_size=0, use_low_rank=False):
    """
    It reduces the entanglement of the given state, producing an approximation
    to reduce the complexity of the quantum circuit needed to prepare it.
    `https://arxiv.org/abs/2111.03132`_.
    Args:
        state_vector (list):
            A state vector to be approximated by a less complex state.
        max_fidelity_loss (float):
            Maximum fidelity loss allowed to the approximated state.
        strategy (string):
            Method to search for the best approximation ('brute_force' or 'greedy').
            For states larger than 2**8, the greedy strategy should preferably be used.
            Default is strategy='greedy'.
        max_combination_size (int):
            Maximum size of the combination ``C(n_qubits, max_combination_size)``
            between the qubits of an entangled subsystem of length ``n_qubits`` to
            produce the possible bipartitions
            (1 <= ``max_combination_size`` <= ``n_qubits``//2).
            For example, if ``max_combination_size``==1, there will be ``n_qubits``
            bipartitions between 1 and ``n_qubits``-1 qubits.
            The default value is 0 (the size will be maximum for each level).
        use_low_rank (bool):
            If set to True, ``rank``>1 approximations are also considered. This is fine
            tuning for high-entanglement states and is slower.
            The default value is False.
    Returns:
        Node: a node with the data required to build the quantum circuit.
    """

    n_qubits = _to_qubits(len(state_vector))

    # Completely separates the state to estimate the maximum possible fidelity loss.
    # If max_fidelity_loss input is lower than the estimated loss, it runs the full
    # routine with potentially exponential cost.
    entanglement, product_state = geometric_entanglement(state_vector, return_product_state=True)
    if max_fidelity_loss >= entanglement:
        qubits = [(n,) for n in range(n_qubits)]
        ranks = [1] * n_qubits
        partitions = [None] * n_qubits
        return Node(
            0, 0, entanglement, entanglement, product_state, qubits, ranks, partitions, []
        )

    vectors = [state_vector]
    qubits = [tuple(range(n_qubits))]
    ranks = [0]
    partitions = [None]

    root_node = Node(0, 0, 0.0, 0.0, vectors, qubits, ranks, partitions, [])
    _build_approximation_tree(root_node, max_fidelity_loss, strategy,
                                    max_combination_size, use_low_rank)

    leaves = []
    _search_leaves(root_node, leaves)

    best_node = _search_best(leaves)

    return best_node

@dataclass
class Entanglement:
    """
    Entanglement reduction information.

    This class contains the information about the entanglement reduction
    of a bipartition. It can be used to assemble an approximate state
    (rank>1) or two completely separate states (rank=1).
    """
    rank: int
    svd_u: np.ndarray
    svd_v: np.ndarray
    svd_s: np.ndarray

    register: Tuple[int]
    partition: Tuple[int]
    local_partition: Tuple[int]

    fidelity_loss: float

@dataclass
class Node:
    """
    Tree node used in _approximation_tree function.
    """
    node_saved_cnots: int
    total_saved_cnots: int

    node_fidelity_loss: float
    total_fidelity_loss: float

    vectors: List[List[complex]]
    qubits: List[Tuple[int]]
    ranks: List[int]
    partitions: List[Optional[Tuple[int]]]

    nodes: List['Node']

    @property
    def is_leaf(self) -> bool:
        """
        True if the all vectors have reached an approximation assessment. There
        is no more decomposition/approximation possible. Therefore, the node is
        a leaf.
        """
        return all(np.asarray(self.ranks) >= 1)

    def num_qubits(self) -> int:
        """ Complete state number of qubits. """
        return len([e for qb_list in self.qubits for e in qb_list])

    def state_vector(self) -> np.ndarray:
        """ Complete state vector. """
        state = ty.tenalg.kronecker(self.vectors) # pylint: disable=no-member
        return state

    def __str__(self):
        str_vectors = '\n'.join([str(np.around(i,2)) for i in self.vectors])
        str_qubits = ' '.join([str(i) for i in self.qubits])
        str_ranks = ' '.join([str(i) for i in self.ranks])
        return f'saved cnots node={self.node_saved_cnots} ' + \
               f'total={self.total_saved_cnots}\n' + \
               f'fidelity loss node={round(self.node_fidelity_loss,6)} ' + \
               f'total={round(self.total_fidelity_loss,6)}\n' + \
               f'states\n{str_vectors}\n' + \
               f'qubits\n{str_qubits}\n' + \
               f'ranks\n{str_ranks}'

def _build_approximation_tree(node, max_fidelity_loss, strategy='brute_force', max_k=0,
                                                                    use_low_rank=False):
    # Ignore states that are already completely or partially disentangled.
    node_data = [(q, v) for q, v, k in zip(node.qubits, node.vectors, node.ranks) if k == 0]

    for entangled_qubits, entangled_vector in node_data:

        if not 1 <= max_k <= len(entangled_qubits)//2:
            max_k = len(entangled_qubits)//2

        if strategy == 'greedy':
            combs = _greedy_combinations(entangled_vector, entangled_qubits, max_k)
        else:
            combs = _all_combinations(entangled_qubits, max_k)

        # Disentangles or reduces the entanglement of each bipartion of
        # entangled_qubits.
        for partition in combs:
            # Computes the two state vectors after disentangling "partition".
            # If the bipartition cannot be fully disentangled, an approximate
            # state is returned.
            entanglement_info = _reduce_entanglement(
                entangled_vector, entangled_qubits, partition, use_low_rank
            )

            node_fidelity_loss = np.array([e_info.fidelity_loss for e_info in entanglement_info])
            total_fidelity_loss = 1.0 - (1.0 - node_fidelity_loss) * \
                                        (1.0 - node.total_fidelity_loss)

            for e_info, loss in zip(entanglement_info, total_fidelity_loss):
                # Recursion should not continue for this branch if
                # "total_fidelity_loss" has reached "max_fidelity_loss".
                # The leaf corresponds to the node of the best approximation of
                # "max_fidelity_loss" on the branch.
                if loss <= max_fidelity_loss:
                    new_node = _create_node(node, e_info)
                    if new_node.total_saved_cnots > 0:
                        node.nodes.append(new_node)

    if len(node.nodes) > 0:  # If it is not the end of the recursion,
        node.vectors.clear() # clear vectors and qubits to save memory.
        node.qubits.clear()  # This information is no longer needed from this point
                             # on (but may be needed in the future).
    if strategy == 'greedy' and len(node.nodes) > 0:
        # Locally optimal choice at each stage.
        node.nodes = [_search_best(node.nodes)]

    for new_node in node.nodes:
        # call _build_approximation_tree recurrently for each new node.
        if not new_node.is_leaf: # Saves one call for each leaf node.
            _build_approximation_tree(
                new_node, max_fidelity_loss, strategy, max_k, use_low_rank
            )

def _all_combinations(entangled_qubits, max_k):
    combs = tuple(combinations(entangled_qubits, max_k))
    if len(entangled_qubits)%2 == 0 and len(entangled_qubits)//2 == max_k:
        # Ignore redundant complements. Only when max_k is exactly
        # half the length of entangled_qubits. Reduces the number of branches.
        # (0,1,2,3) -> (0,1), (0,2), (0,3); ignore (2,3), (1,3), (1,2) .
        combs = combs[:len(combs)//2]

    return chain(*(combinations(entangled_qubits, k)
                                            for k in range(1, max_k)), combs)

def _greedy_combinations(entangled_vector, entangled_qubits, max_k):
    """
    Combinations with a qubit-by-qubit analysis.
    Returns only one representative of the partitions of size k (1<=k<=max_k).
    The increment in the partition size is done by choosing the qubit that has
    the lowest fidelity-loss when removed from the remaining entangled subsystem.
    """
    node = Node( 0, 0, 0.0, 0.0, [entangled_vector], [entangled_qubits], [0], [None], [] )
    for _ in range(max_k):
        current_vector = node.vectors[-1] # Last item is the current entangled state.
        current_qubits = node.qubits[-1]

        nodes = []
        # Disentangles one qubit at a time.
        for qubit_to_disentangle in current_qubits:
            entanglement_info = \
                _reduce_entanglement(current_vector, current_qubits, (qubit_to_disentangle,))

            new_node = _create_node(node, entanglement_info[0])

            nodes.append(new_node)
        # Search for the node with lowest fidelity-loss.
        node = _search_best(nodes)

    # Build the partitions by incrementing the number of selected qubits.
    # Returns only one partition for each length k.
    # All disentangled qubits are in the slice "node.qubits[0:max_k]", in the order in which
    # they were selected. Each partition needs to be sorted to ensure that the correct
    # construction of the circuit.
    return ( tuple(sorted( chain(*node.qubits[:k]) )) for k in range(1, max_k+1) )

def _reduce_entanglement(state_vector, register, partition, use_low_rank=False):
    local_partition = []
    # Maintains the relative position between the qubits of the two subsystems.
    for qubit_to_disentangle in partition:
        local_partition.append(sum(i < qubit_to_disentangle for i in register))

    local_partition = tuple(local_partition)

    svd_u, svd_s, svd_v = schmidt_decomposition(state_vector, local_partition)

    entanglement_info = []

    max_ebits = 0
    if use_low_rank:
        # Limit the maximum low_rank to "2**(total_ebits-1)" to not repeat the original state.
        max_ebits = _to_qubits(svd_s.shape[0]) - 1

    for ebits in range(0, max_ebits+1):
        low_rank = 2**ebits

        rank, low_rank_u, low_rank_v, low_rank_s = \
            low_rank_approximation(low_rank, svd_u, svd_v, svd_s)

        if rank < low_rank:
            break # No need to go any further, as the maximum effective rank has been reached.

        fidelity_loss = 1.0 - sum(low_rank_s**2)

        entanglement_info.append(Entanglement(rank, low_rank_u, low_rank_v, low_rank_s,
                                              register,
                                              partition,
                                              local_partition,
                                              fidelity_loss))
    return entanglement_info

def _create_node(parent_node, e_info):

    vectors = parent_node.vectors.copy()
    qubits = parent_node.qubits.copy()
    ranks = parent_node.ranks.copy()
    partitions = parent_node.partitions.copy()

    index = parent_node.qubits.index(e_info.register)
    original_vector = vectors.pop(index)
    original_qubits = qubits.pop(index)
    original_rank = ranks.pop(index)
    original_partition = partitions.pop(index)

    if e_info.rank == 1:
        # The partition qubits have been completely disentangled from the
        # rest of the register. Therefore, the original entangled state is
        # removed from the list and two new separate states are included.
        partition1 = tuple( sorted(set(original_qubits).difference(set(e_info.partition))) )
        partition2 = e_info.partition

        vectors.append(e_info.svd_v.T[:, 0])
        qubits.append(partition2)
        ranks.append(1 if len(partition2) == 1 else 0) # Single qubit states can
        partitions.append(None)                        # no longer be disentangled.

        vectors.append(e_info.svd_u[:, 0])
        qubits.append(partition1)
        ranks.append(1 if len(partition1) == 1 else 0)
        partitions.append(None)

        node_saved_cnots = _count_saved_cnots(
                                original_vector, vectors[-1], vectors[-2],
                                original_partition, None,
                                original_rank
                            )
    else:
        # The entanglement between partition qubits and the rest of the
        # register has been reduced, but not eliminated. Therefore, the
        # original state is replaced by an approximate state.
        normed_svd_s = e_info.svd_s/sqrt( 1.0 - e_info.fidelity_loss )
        approximate_state = schmidt_composition(e_info.svd_u, e_info.svd_v,
                                                normed_svd_s, e_info.local_partition)
        vectors.append(approximate_state)
        qubits.append(original_qubits)
        ranks.append(e_info.rank)
        partitions.append(e_info.local_partition)

        node_saved_cnots = _count_saved_cnots(
                                original_vector, vectors[-1], None,
                                original_partition, e_info.local_partition,
                                original_rank, e_info.rank
                            )

    total_saved_cnots = parent_node.total_saved_cnots + node_saved_cnots
    total_fidelity_loss = 1.0 - (1.0 - e_info.fidelity_loss) * \
                                (1.0 - parent_node.total_fidelity_loss)

    return Node(node_saved_cnots, total_saved_cnots, e_info.fidelity_loss,
                    total_fidelity_loss, vectors, qubits, ranks, partitions, [])

def _search_leaves(node, leaves):
    # It returns the leaves of the tree. These nodes are the ones with
    # total_fidelity_loss closest to max_fidelity_loss for each branch.
    if len(node.nodes) == 0:
        leaves.append(node)
    else:
        for child in node.nodes:
            _search_leaves(child, leaves)

def _search_best(nodes):
    # Nodes with the greatest reduction in the number of CNOTs.
    # There may be several with the same number.
    max_total_saved_cnots = max(nodes, key=lambda n: n.total_saved_cnots).total_saved_cnots
    max_saved_cnots_nodes = [node for node in nodes
                                if node.total_saved_cnots == max_total_saved_cnots]
    # Nodes with the minimum depth (wich depends on the size of the node's largest subsystem).
    # Shallower circuits with the same number of CNOTs means more parallelism.
    min_depth = _max_subsystem_size(min(max_saved_cnots_nodes, key=_max_subsystem_size))
    min_depth_nodes = [node for node in max_saved_cnots_nodes
                                if _max_subsystem_size(node) == min_depth]
    # Node with the lowest fidelity loss among the nodes with
    # the highest reduction in the number of CNOTs.
    return min(min_depth_nodes, key=lambda n: n.total_fidelity_loss)

def _max_subsystem_size(node):
    return len(max(node.qubits, key=len))

def _to_qubits(n_state_vector):
    return int(log2(n_state_vector))

def _count_saved_cnots(original_vector, subsystem1_vector, subsystem2_vector,
                            original_partition=None,subsystem_local_partition=None,
                            original_rank=0, subsystem_rank=0):
    method = 'estimate'

    cnots_originally = schmidt_cnots(
                        original_vector, method=method, partition=original_partition,
                        low_rank=original_rank
                    )
    cnots_phase_3 = schmidt_cnots(
                        subsystem1_vector, method=method, partition=subsystem_local_partition,
                        low_rank=subsystem_rank
                    )

    cnots_phase_4 = 0
    if subsystem2_vector is not None:
        cnots_phase_4 = schmidt_cnots(subsystem2_vector, method=method)

    return cnots_originally - cnots_phase_3 - cnots_phase_4



"""
pypi setup
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="qclib",
    version="0.0.9",
    author="Adenilton Silva",
    author_email="ajsilva@cin.ufpe.br",
    description="A quantum computing library using qiskit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/qclib/qclib",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'scipy>=1.7.1',
        'qiskit>=0.18.3',
        'deprecation',
        'graphviz'
    ]
)



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
https://arxiv.org/abs/quant-ph/9807054
"""

import numpy as np
from qiskit import QuantumCircuit, QuantumRegister

# pylint: disable=maybe-no-member

def initialize(state, n_qubits, n_output_values):
    """ State preparation using Ventura and Martinez algorithm quant-ph/9807054
    Algorithm that requires a polynomial number of elementary operations for
    initializing a quantum system to represent only the m known points of a
    function f (m = len(state)).
    The result is a quantum superposition with m nonzero coefficients -- the
    creation of which is a nontrivial task compared to creating a superposition
    of all basis states.
    The amplitudes of modulus "1/sqrt(m)" will be "2 pi / N" radians apart from
    each other on the complex plane.

    Binary output function case:
        f:z->s with z \\in {0,1}^n and s \\in {0, 1}
    General case:
        f:z->s with z \\in {0,1}^n and s \\in {0, 1, ..., N-1}

    For instance, to initialize the state
    1/sqrt(3)|01> + 1/sqrt(3)*e^(1*i*2pi/N)|10> + 1/sqrt(3)*e^(2*i*2pi/N)c|11>
        $ state = {1: 0, 2: 1, 3: 2}
        $ circuit = initialize(state, n=2, N=3)

    Parameters
    ----------
    state: dict of {int:float}
        A unit vector representing a quantum state.
        Keys are function binary input values and values are function output values.

    n_qubits: int
        Number of bits to represent the binary values.

    n_output_values: int
        Number of possible output values N (Ex.: n_output_values=2 for a binary function).

    Returns
    -------
    circuit: QuantumCircuit
        QuantumCircuit to initialize the state.
    """

    reg_x = QuantumRegister(n_qubits, 'x')
    reg_g = QuantumRegister(n_qubits - 1, 'g')
    reg_c = QuantumRegister(2, 'c')

    circuit = QuantumCircuit(reg_x, reg_g, reg_c)

    reg_x = reg_x[::-1] # qiskit reverse (qiskit little-endian)

    bits_z0 = [int(k) for k in '{:0{}b}'.format(0, n_qubits)]
    for idx_p, (input_z, output_s) in list(enumerate(state.items()))[::-1]:
        bits_z = [int(k) for k in '{:0{}b}'.format(input_z, n_qubits)]

        circuit.x(reg_c[1])
        for j,_ in enumerate(bits_z):
            if bits_z0[j] != bits_z[j]:
                circuit.cx(reg_c[1], reg_x[j])

        bits_z0 = bits_z
        circuit.cx(reg_c[1], reg_c[0])
        circuit.x(reg_c[1])

        _apply_smatrix(circuit, idx_p, n_output_values, output_s, reg_c)

        _flipflop01(bits_z, circuit, reg_x)

        circuit.ccx(reg_x[0], reg_x[1], reg_g[0])

        _flipflop01(bits_z, circuit, reg_x)

        for k in range(2, n_qubits):
            if bits_z[k] == 0:
                circuit.x(reg_x[k])

            circuit.ccx(reg_x[k], reg_g[k-2], reg_g[k-1])

            if bits_z[k] == 0:
                circuit.x(reg_x[k])

        circuit.cx(reg_g[n_qubits - 2], reg_c[0])

        for k in range(n_qubits - 1, 1, -1):
            if bits_z[k] == 0:
                circuit.x(reg_x[k])

            circuit.ccx(reg_x[k], reg_g[k-2], reg_g[k-1])

            if bits_z[k] == 0:
                circuit.x(reg_x[k])

        _flipflop01(bits_z, circuit, reg_x)

        circuit.ccx(reg_x[0], reg_x[1], reg_g[0])

        _flipflop01(bits_z, circuit, reg_x)

    circuit.x(reg_c[1])

    return circuit


def _apply_smatrix(circuit, idx_p, n_output_values, output_s, reg_c):
    theta = -2 * np.arccos(np.sqrt(idx_p / (idx_p + 1)))  # This sign is here for the smaller
                                                          # values of "s" to be represented by
                                                          # negative amplitudes and the larger
                                                          # ones by positive amplitudes.
    lamb = -output_s * 2 * np.pi / n_output_values        # In the paper this negative sign is
                                                          # missing. Without it the matrix S
                                                          # is not unitary.
    phi = -lamb
    circuit.cu(theta, phi, lamb, 0, reg_c[0], reg_c[1])


def _flipflop01(bits_z, circuit, reg_x):
    if bits_z[0] == 0:
        circuit.x(reg_x[0])
    if bits_z[1] == 0:
        circuit.x(reg_x[1])



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the isometry.py module.
"""

from unittest import TestCase
import numpy as np
from qiskit import QuantumCircuit
from scipy.stats import unitary_group
from qclib.isometry import decompose
from qclib.util import get_state

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

class TestIsometry(TestCase):

    # scheme='knill' isometry tests

    def test_state_preparation_real_knill(self):
        self._test_state_preparation_real('knill')

    def test_state_preparation_complex_knill(self):
        self._test_state_preparation_complex('knill')

    def test_isometry_knill(self):
        self._test_isometry('knill')

    def test_fixed_isometry_knill(self):
        self._test_fixed_isometry('knill')

    # scheme='ccd' (column-by-column decomposition) isometry tests

    def test_state_preparation_real_ccd(self):
        self._test_state_preparation_real('ccd')

    def test_state_preparation_complex_ccd(self):
        self._test_state_preparation_complex('ccd')

    def test_isometry_ccd(self):
        self._test_isometry('ccd')

    def test_fixed_isometry_ccd(self):
        self._test_fixed_isometry('ccd')

    # general functions for testing the schemes (knill and ccd).

    def _test_state_preparation_real(self, scheme):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = decompose(state_vector, scheme=scheme)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state))

    def _test_state_preparation_complex(self, scheme):
        state_vector = np.random.rand(32) + np.random.rand(32) * 1j
        state_vector = state_vector / np.linalg.norm(state_vector)

        circuit = decompose(state_vector, scheme=scheme)

        state = get_state(circuit)

        self.assertTrue(np.allclose(state_vector, state))

    def _test_isometry(self, scheme):
        log_lines = 4
        log_cols = 2
        isometry = unitary_group.rvs(2**log_lines)[:,:2**log_cols]

        gate = decompose(isometry, scheme=scheme)

        for j in range(2**log_cols):
            circuit = QuantumCircuit(log_lines)

            for i, bit in enumerate('{:0{}b}'.format(j, 1)[::-1]):
                if bit == '1':
                    circuit.x(i)

            circuit.append(gate, circuit.qubits)
            state = get_state(circuit)

            self.assertTrue(np.allclose(isometry[:, j], state))

    def _test_fixed_isometry(self, scheme):
        isometry = np.copy([[-0.5391073,  -0.12662419, -0.73739705, -0.38674956],
                            [ 0.15705405,  0.20566939,  0.32663193, -0.9090356 ],
                            [-0.77065035, -0.23739918,  0.59084039,  0.02544217],
                            [-0.30132273,  0.9409081,  -0.02155946,  0.15307435]])

        gate = decompose(isometry, scheme=scheme)

        state = get_state(gate)

        self.assertTrue(np.allclose(isometry[:, 0], state))

        circuit = QuantumCircuit(2)
        circuit.x(0)
        circuit.append(gate, circuit.qubits)
        state = get_state(circuit)
        self.assertTrue(np.allclose(isometry[:, 1], state))

        circuit = QuantumCircuit(2)
        circuit.x(1)
        circuit.append(gate, circuit.qubits)
        state = get_state(circuit)
        self.assertTrue(np.allclose(isometry[:, 2], state))

        circuit = QuantumCircuit(2)
        circuit.x([0,1])
        circuit.append(gate, circuit.qubits)
        state = get_state(circuit)
        self.assertTrue(np.allclose(isometry[:, 3], state))



# Copyright 2021 qclib project.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
qclib auxiliary functions
"""

from qiskit import execute, transpile
from qiskit.providers.aer.backends import AerSimulator
import numpy as np
from scipy import sparse


def get_counts(circ):
    """
    Parameters
    ----------
    circ: QuantumCircuit (with measurement operations)

    Returns
    -------
    counts: output generated by the quantum circuit
    """
    backend = AerSimulator()
    counts = execute(circ, backend).result().get_counts()
    return counts


def get_state(circ):
    """
    Parameters
    ----------
    circ: QuantumCircuit

    Returns
    -------
    state_vector: state generated by the quantum circuit
    """

    backend = AerSimulator()
    tcirc = transpile(circ, backend)
    tcirc.save_statevector()
    state_vector = backend.run(tcirc).result().get_statevector()

    return np.array(state_vector)


def replace_all_values_with(new_value, dataset):
    """
        Given a list of tuples (v, b),where v is the value
        and b is the binary pattern associated to it.
        this procedure performs the task of replacing
        v with the new_value
    :param new_value: Value to replate the v in all the tuples
                      (v, b)
    :param dataset: List of tuples where the values are to be
                    replaced
    :return: new list of tuples
    """

    new_dataset = []
    for _, binary_pattern in dataset:
        new_dataset.append((new_value, binary_pattern))

    return new_dataset


def build_list_of_quibit_objects(quantum_register):
    """
        Buid a list of Qubit objects to be used as
        input to some procedure of the qiskit framework
    :param quantum_register: Quantum register with the qubits
    :return: Qubits list
    """
    qubits_list = []

    for i in range(quantum_register.size):
        qubits_list.append(quantum_register[quantum_register.size - i - 1])

    return qubits_list


def verify_interval_in_state_vector(statevector, start, finish):
    """
        Verifies if there is at least one non zero entry in
        a given interval in the state vectors cells, and
        returns true if positive
    :param statevector: state vector to be processed
    :param start: start of the interval
    :param finish: end of the interval
    :return: Boolean True if a non zero entry has been found
    """
    found = False
    for cell_idx in range(start, finish):

        cell_value = statevector[cell_idx]
        if cell_value != 0:
            found = True
            break
    return found


def verify_trigonometric_interval(value):
    """
        Verify if a certain value is inside the interval
        of the domain of the tirgonometric functions
        cosine and sine, [-1, 1]
    :param value: Real value to be evaluated
    :return: Value, if the value is inside the domain
             Updated value, if the value is outside the
             domain
    """

    value = min(value, 1)
    if value < -1:
        value = -1
    return value


def _count_ones(pattern):
    return pattern[0].count(1)


def random_sparse(nbits, density):
    '''
    Creates a random input for sparse quantum state preparation
    nbits: int number of qubits
    density: float in [0,1]

    returns
    bin_data: [(binary_string_k, float_k)] k = 0 ... n
    '''

    data = sparse.random(2 ** nbits, 1, density, format="dok")

    rows, _ = data.nonzero()
    bin_data = []

    length = sparse.linalg.norm(data)

    for k in rows:
        bin_data.append((format(k, "0" + str(nbits) + "b"), data[k, 0] / length))

    bin_data.sort(key=_count_ones)
    return bin_data

def _double_sparse_binary(nbits, log_size, p_1, p_0):

    bin_data = []
    while len(bin_data)< 2**log_size:
        lst = np.random.choice(2, nbits, p=[p_1, p_0]).tolist()

        if lst not in bin_data:
            bin_data.append(lst)

    return  bin_data


def double_sparse(nbits, log_size, p_1):
    """
    Parameters
    ----------
    nbits (int): number of qubits
    log_size (int): log_2(number of amplitudes)
    p_1 (float): probability of qubit equal to one

    Returns
    -------
    \\sum_{k} x_k |p_k>, each bit of p_k is equal to 1 with probability p1
    """
    data = np.random.rand(2 ** log_size) + np.random.rand(2 ** log_size)*1j
    length = np.linalg.norm(data)
    data = (1/length) * data

    binary = _double_sparse_binary(nbits, log_size, 1 - p_1, p_1)
    bin_data = [(binary[i], data[i]) for i in range(2 ** log_size)]

    bin_data.sort(key=_count_ones)
    return bin_data


def _compute_matrix_angles(feature, norm):
    """
        Compute the angles of the matrix U3 necessary for encoding
        the phase of the state
    :param feature: Complex or float, feature to be stored
    :param norm: remaining norm to be used to compute the angles
    :return: the angles alpha, beta and phi of the operator U3
    """
    alpha = 0
    beta = 0
    phi = 0

    if isinstance(feature, complex):
        phase = np.abs(np.power(feature, 2))

        if (norm - phase) < 0:
            norm = phase

        cos_value = np.sqrt((norm - phase) / norm)
        cos_value = verify_trigonometric_interval(cos_value)
        alpha = 2 * (np.arccos(cos_value))
        beta = np.arccos(- feature.real / np.sqrt(np.abs(np.power(feature, 2))))

        if feature.imag < 0:
            beta = 2 * np.pi - beta

        phi = - beta

    else:
        sin_value = - feature / np.sqrt(norm)
        sin_value = verify_trigonometric_interval(sin_value)
        alpha = 2 * (np.arcsin(sin_value))

    return alpha, beta, phi
